/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf;

/**
 * Layered architecture allows an efficient processing of messages
 * passed by reference from one <em>layer</em> to another.
 * Every layer looks at the message and evaluates the information
 * that it is capable to understand.
 * A layer can call other  layers to
 * accomplish its work.
 * <p>Example of layer initialization:
 * <p><!hr><blockquote><pre>
 *    Layer service = new MyService(factory);
 *    Layer transport = new HTTPLayer(factory);
 *    service.connect( "transport", transport );
 *    transport.connect( "dispatch", service );
 * </pre></blockquote><!hr><!p>
 *
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public interface Layer {

  /**
   * A processing request for the layer.
   * Model can be modified within this call, so
   * call dispatch(msg.duplicate(), from) to ensure
   * the original model remains intact
   *
   * @param      msg   the message to dispatch
   * @param      from  which Layer the message came from
   *
   * @return Model (optional, null if none)
   */
  public Model dispatch(Model msg, Layer from) throws Exception;

  /*
   * Some standard way of connecting layers.
   *
   * @param      what   the channel ID as a string
   * @param      to     the layer which is to connect to the given channel
   *
     public void connect(String what, Layer to) throws Exception;
   */
}
