/**
 * Copyright  Sergey Melnik (Stanford University, Database Group)
 *
 * All Rights Reserved.
 */

package org.ginf.helpers;

import org.ginf.*;
import org.w3c.rdf.*;
import java.util.*;

import org.ginf.schemas.RDF;
import org.ginf.schemas.RDFS;

/**
 * This class implements evaluation of RDF schema information.
 * This default implementation works only with
 * the default implementation of the registry: SchemaRegistryImpl
 *
 * @author Sergey Melnik <melnik@db.stanford.edu>
 */

public class SchemaModelImpl implements SchemaModel {
  
  Model peer;
  SchemaRegistryImpl registry;

  /**
   * Works only with SchemaRegistryImpl!
   * Peer model must be a flat model.
   */
  public SchemaModelImpl(SchemaRegistryImpl registry, Model peer) {
    this.registry = registry;
    this.peer = peer;
    if(peer == null)
      throw new IllegalArgumentException("[SchemaModel] invalid peer: null");
  }

  /**
   * Enumeration of triples.
   */
  public Enumeration elements() {
    return peer.elements();
  }

  /**
   * Number of triples.
   */
  public int size() {
    return peer.size();
  }

  /**
   * Whether contains given triple
   */
  public boolean contains(Triple t) {
    return peer.contains(t);
  }

  // Model manipulation: add, remove, find

  /**
   * Adds a new triple to the message. Object is a String (literal).
   */
  public void add(Resource subject, Resource predicate, String object) {
    peer.add( subject, predicate, object );
  }

  /**
   * Adds a new triple to the message. Object is an RDFnode.
   */
  public void add(Resource subject, Resource predicate, RDFnode object) {
    peer.add( subject, predicate, object );
  }

  public void add(Triple t) {
    peer.add(t);
  }

  /**
   * Removes the triple from the message.
   */
  public void remove(Triple t) {
    peer.remove(t);
  }

  /**
   * Cloning
   */
  public Model duplicate() {

    return new SchemaModelImpl(registry, peer.duplicate());
  }

  public Model create() {
    return new SchemaModelImpl(registry, peer.create());
  }

  /**
   * Set union with another model
   */
  public Model union(Model m) {
    peer.union( m );
    return this;
  }

  /**
   * Set difference with another model
   */
  public Model difference(Model m) {
    peer.difference( m );
    return this;
  }

  public void setURI(String uri) {
    peer.setURI(uri);
  }

  public String getURI() {
    return peer.getURI();
  }

  //  public Model find( Resource subject, Resource predicate, RDFnode object ) {

    // ensure we not enter a dead loop
  //    return find( subject, predicate, object, 0 );
  //  }

  //  static final int DEAD_LOOP = 30; // recursive subClassOf ?

  /**
   * Loop counter is passed to provide a simple safety mechanism.
   */

  /**
   * Search for triples.
   * NULL input parameter match anything.
   * Returns a flat model as a result!
   */
  public Model find( Resource subject, Resource predicate, RDFnode object /*, int loop*/ ) {

    //    if(loop > DEAD_LOOP)
    //      throw new RuntimeException("[SchemaModel] dead loop encountered, contains edge: (" +
    //				 subject + ", " + predicate + ", " + object);

    Model closure = registry.getClosure();

    if(predicate != null &&
       ( predicate.equals(RDFS.subClassOf) ||
	 predicate.equals(RDFS.subPropertyOf )) ) {

      return closure.find( subject, predicate, object );
    }

    if( RDF.type.equals(predicate) ) {

      //   (type)
      //   X rdf:type SUPERCLASS :- SUBCLASS rdfs:subClassOf SUPERCLASS && X rdf:type SUBCLASS
      
      // example: null rdf:type DL:Search not there,
      // but X rdf:type S:SpecSearch && S:SpecSearch rdfs:subClassOf DL:Search

      // FIXME: what about rdfs:Literals? Consider: X rdf:type rdfs:Literal
      // after W3C is confident about Literals!
      // So far:

      // special treatment for Literals
      if( RDFS.Literal.equals( object ) ) {

	//	System.err.println("SPEC:LITERALS");

	// collect all explicit literals
	Model m = peer.find( subject, RDF.type, object );
	m.union( registry.get().find( subject, RDF.type, RDFS.Literal ) );

	Model added = peer.find ( subject, null, null );
	added.union( registry.get().find( subject, null, null ) );
	// now filter all literals we've found
	for(Enumeration en_l = added.elements(); en_l.hasMoreElements();) {
	  Triple t_a = (Triple)en_l.nextElement();
	  if( t_a.object() instanceof Literal )
	    m.add( t_a );
	}
	return m;

      } else if(RDFS.Resource.equals(object)) {

	// special treatment of resources
	Model res = peer.create();
	if(subject != null)
	  res.add( subject, RDF.type, RDFS.Resource );
	else {
	  // collect all resources
	  Hashtable done = new Hashtable(); // resource -> [Triple OR RDFS.Resource]
	  for(Enumeration en = peer.elements(); en.hasMoreElements(); ) {
	    
	    Triple t_r = (Triple)en.nextElement();
	    Resource r = t_r.subject();

	    if(RDF.type.equals(t_r.predicate()))
	      done.put( r, t_r ); // always overwrite
	    else if(done.get(r) == null)
	      done.put( r, RDFS.Resource );
	    if( t_r.object() instanceof Resource && done.get(r) == null )
	      done.put( r, RDFS.Resource );
	  }

	  // add to the model
	  for(Enumeration en = done.keys(); en.hasMoreElements(); ) {

	    Resource r = (Resource)en.nextElement();
	    Object v = done.get(r);
	    if(v instanceof Triple)
	      res.add((Triple)v);
	    else
	      res.add( r, RDF.type, RDFS.Resource );
	  }
	}
	return res;

      } else {

	// we are NOT looking for literal types

	Model res = peer.create();

	// Ordinal numbers, ContainerMembershipProperty
	// FIXME: infinite result set, do something?
	if(RDFUtil.getOrd(subject) > 0 &&
	   (object == null || RDFS.ContainerMembershipProperty.equals(object))) {
	  res.add( subject, RDF.type, RDFS.ContainerMembershipProperty );
	  return res;
	}

	Model m = peer.find( subject, RDF.type, null );
	// FIXME: merge with schema information???

	/*
	if(m.size() == 4 && subject == null && object != null && object.toString().endsWith("Connection")) {
	  System.err.println("BAD PEER: " + peer);
	  //	  RDFUtil.printTriples( peer, System.out );
	  //	  (Hashtable)peer.objects.get(RDF.type)
	}
	*/

	//	m.union( registry.get().find( subject, RDF.type, null ) );

	//	System.out.println("SPECIAL:TYPE so far:" + m.size() + " for " + subject + ", " + object);
	//	RDFUtil.printTriples( m, System.out );
      
	for(Enumeration en = m.elements(); en.hasMoreElements();) {
	  
	  Triple t = (Triple)en.nextElement();
	  Resource subClass = (Resource)t.object();

	  // if subClass rdfs:subClassOf object take this triple
	  if( object != null ) {
	    // search for existing triples
	    //	    System.out.println("COMPARING " + subClass + " TO " + object + ": " + subClass.equals(object));

	    if( subClass.equals(object) ||
		closure.find( subClass, RDFS.subClassOf, object ).size() > 0 )
	      res.add( t );
	  } else {
	    // search for all superclasses
	    res.add( subject, RDF.type, subClass );
	    Model sc = closure.find( subClass, RDFS.subClassOf, null );
	    for(Enumeration en_sc = sc.elements(); en_sc.hasMoreElements();) {
	      res.add( subject, RDF.type, ((Triple)en_sc.nextElement()).object() );
	    }
	  }
	}

	return res;
      }
    }

    // special case: predicate might be rdfs:subPropertyOf something
    // X SUBPROP Y :- SUBPROP rdfs:subPropertyOf SUPERPROP && X SUPERPROP Y

    // optimized recursive search
	  
    Model m = peer.find( subject, predicate, object );
    Model temp = closure.find( null, RDFS.subPropertyOf, predicate );

    if(temp.size() > 0) {
      for(Enumeration en_t = temp.elements(); en_t.hasMoreElements(); ) {
	Resource subProp = (Resource)((Triple)en_t.nextElement()).subject();
	Model added = peer.find( subject, subProp, object );
	//	System.err.println("<ADDING: " + subProp + "," + object + ": " + added.size());
	//	RDFUtil.printTriples( added, System.err);
	//	System.err.println("ADDING>");
	m.union ( added );
      }
    }

    //    System.out.println("REGISTRY AGAIN:");
    //    DefaultModel.printTriples(registry.get(), System.out);

    //    System.err.println("SUBPROPERTIES of " + predicate + " found: " + temp.size());
    /*
    if(temp.size() > 0) {
      for(Enumeration en_t = temp.elements(); en_t.hasMoreElements(); ) {
	// a property can be a resource, too
	Resource subProp = (Resource)((Triple)en_t.nextElement()).subject();
	Model added = find( subject, subProp, object, loop+1 );
	m.union ( added );
      }
    }
    */
    
    return m;
  }

    /* 
      // special case: predicate is rdfs:subClassOf
      // FIXME: treat subPropertyOf exactly as subClassOf
      // we do not even look in the flat model, but in the registry

      // peer is flat, returns flat m
      Model m = registry.get().find( subject, registry.subClassOf, object );

      //      System.err.println("SPECIAL:SUBCLASS, " + m.size() + " flat records for " + subject + ", " + object);

      // X rdfs:subClassOf Y :- X rdfs:subClassOf Z && Z rdfs:subClassOf Y
      // transitive closure!
      // null subClassOf null -> all pairs
      // null subClassOf B    -> pairs (X, B)
      // A subClassOf null    -> pairs (A, X)
      
      // cut: (A, X) and nothing found -> done
      if(object == null && m.size() == 0) {
	//	if(m.size() == 0)
	return new SchemaModel( registry, m );

      } else {

	// (A, X), something found. Find again (Xi, Y)
	for(Enumeration en = m.elements(); en.hasMoreElements();) {
	  
	  Triple t = (Triple)en.nextElement();
	  Resource Xi = (Resource)t.object();
	  // call myself recursively
	  Model again = find( Xi, registry.subClassOf, null, loop+1 );
	  // add triples
	  for(Enumeration en_super = again.elements(); en_super.hasMoreElements();) {
	    
	    Triple t_super = (Triple)en_super.nextElement();
	    Resource superR = (Resource)t_super.object();
	    m.add ( t.subject(), registry.subClassOf, superR );
	  }
	}
	return new SchemaModel( registry, m );
      }
    }
    */

      /*
      // compute transitive closure on the predicate without recursion

      Model base = registry.get().find( null, predicate, null );
      Model all = base; // registry.get().find( null, predicate, null );

      //      System.err.println("Computing closure for: " + subject + ", " + predicate + ", " + object);

      int old_size;
      do {
	Model added = peer.create();
	
	for(Enumeration en_p = base.elements(); en_p.hasMoreElements();) {
	  Triple pivot = (Triple)en_p.nextElement();
	  
	  for(Enumeration en_a = all.elements(); en_a.hasMoreElements();) {
	    Triple t = (Triple)en_a.nextElement();
	    if(t.subject().equals(pivot.object()))
	      added.add( pivot.subject(), predicate, t.object() );
	    if(pivot.subject().equals(t.object()))
	      added.add( t.subject(), predicate, pivot.object() );
	  }
	}
	old_size = base.size();
	base.union(added);
      } while (base.size() > old_size);
      // base is flat
      return base.find( subject, predicate, object );
    }
    */

  /**
   * Search for a single triple.
   * NULL input parameter match anything.
   */
  public Triple find1( Resource subject, Resource predicate, RDFnode object ) {

    Model m = find(subject, predicate, object);

    if(m.size() == 1)
      return (Triple)m.elements().nextElement();
    else
      if(m.size() > 1)
	throw new RuntimeException("Model: multiple triples encountered in find1");
    return null;
  }


  // see FIXED below
  //  static final Resource RDF_Description = new Resource(RDF.NS + "Description");

  /**
   * Validates a model. In case of an invalid statement
   * checks whether schema was not loaded
   */
  public void validate() {

    Hashtable containers = new Hashtable(); // triples containing collections and ordinals
    StringBuffer errors = new StringBuffer();

    // merge schema and model for special rdf:type cases
    SchemaModel regExt = new SchemaModelImpl( registry, registry.get().duplicate().union( peer ) );

    for(Enumeration en = peer.elements(); en.hasMoreElements();) {
	  
      Triple t = (Triple)en.nextElement();

      // rdf:type
      if(RDF.type.equals(t.predicate())) {
	// ensure that the target is of type rdf:Class
	if(t.object() instanceof Literal)
	  invalid( errors, t, "Literals cannot be used for typing" );

	// FIXED: rdf:Descriptions cannot actually be present as triples
	// in case reified Statements are present:
	//	if(RDF_Description.equals(t.object()))
	//	  continue; // just ignore it

	Model res = regExt.find( (Resource)t.object(), RDF.type, RDFS.Class );
	if(res.size() == 0) {
	  //	  System.err.println("SUPERCLASSES(C)");
	  //	  RDFUtil.printTriples( regExt.find( (Resource)t.object(), RDF.type, null ), System.err);
	  if(noSchema(errors, t.object().toString()))
	    break;
	  else
	    invalid( errors, t, t.object().toString() + " must be an instance of " + RDFS.Class);
	}

      } else if(RDFUtil.getOrd(t.predicate()) > 0) {

	// save for later
	containers.put( t.subject(), t );

      } else {
	// check domain and range

	Vector expected = new Vector();

	// find all allowed domains of the Property
	Model domains = registry.get().find( t.predicate(), RDFS.domain, null );
	if(domains.size() > 0) {
	  boolean domainOK = false;
	  // go through all valid domains and check whether
	  // the subject() is an instance of a valid domain Class
	  for(Enumeration en_d = domains.elements(); en_d.hasMoreElements();) {
	    Resource domainClass = (Resource)((Triple)en_d.nextElement()).object();
	    //	    System.err.println("DOMAIN: checking " + t.subject() + " against " + domainClass);
	    expected.addElement( domainClass );
	    if( regExt.find( t.subject(), RDF.type, domainClass ).size() != 0) {
	      //	      System.err.println("DOMAIN CHECKED: " + domainClass);
	      domainOK = true;
	      break;
	    }
	  }
	  if(!domainOK) {
	    //            System.err.println("SUPERCLASSES");
	    //            RDFUtil.printTriples( regExt.find( t.subject(), RDF.type, null ), System.err);
	    if(noSchema(errors, t.subject().toString()))
	      break;
	    else
	      invalid( errors, t, "Bad domain. Expected: " + expected );
	  }
	}

	expected.setSize(0);
	// find all allowed domains of the Property
	Model ranges = registry.get().find( t.predicate(), RDFS.range, null );
	if(ranges.size() == 1) { // there can be only one range property!!! (See specs)
	  boolean rangeOK = false;
	  // go through all valid ranges and check whether
	  // the object() is an instance of a valid range Class
	  for(Enumeration en_r = ranges.elements(); en_r.hasMoreElements();) {
	    Resource rangeClass = (Resource)((Triple)en_r.nextElement()).object();
	    //	    System.err.println("RANGE: checking " + t.object() + " against " + rangeClass);
	    expected.addElement(rangeClass);
	    // special treatment for Literals
	    if(RDFS.Literal.equals(rangeClass)) {
	      if( t.object() instanceof Literal ) {
		//		System.err.println("RANGE CHECKED: Literal");
		rangeOK = true;
		break;
	      } else
		invalid( errors, t, t.object() + " must be a literal");

	    } else if ( t.object() instanceof Resource &&
			regExt.find( (Resource)t.object(), RDF.type, rangeClass ).size() != 0 ) {
	      //	      System.err.println("RANGE CHECKED: " + rangeClass);
	      rangeOK = true;
	      break;
	    }
	  }
	  if(!rangeOK)
	    if(noSchema(errors, t.object().toString()))
	      break;
	    else
	      invalid( errors, t, "Bad range. Expected: " + expected);

	} else if(ranges.size() > 1)
	  invalid( errors, null, "Invalid schema. Multiple ranges for " + t.predicate() );
      }
    }
    
    for(Enumeration en = containers.elements(); en.hasMoreElements();) {
	  
      Triple t = (Triple)en.nextElement();
      Resource c = t.subject();
      // must be subclass of Container
      if( regExt.find( t.subject(), RDF.type, RDFS.Container ).size() == 0 )
	invalid( errors, t, t.subject().toString() + " must be an instance of rdfs:Container" );
    }

    if(errors.length() > 0) {
      //      System.err.println("MODEL");
      //      RDFUtil.printTriples(peer, System.err);
      //      System.err.println("SCHEMA");
      //      RDFUtil.printTriples(registry.get(), System.err);
      throw new RuntimeException("INVALID MODEL: " + errors.toString());
    }
    
  }

  boolean noSchema(StringBuffer b, String uri) {

    if(!registry.contains(RDFUtil.getNS(uri))) {
      invalid( b, null, "No schema available for " + uri + ". Possible reason: API-constructed model, has to fetch schema explicitly");
      return true;
    }
    return false;
  }


  void invalid(StringBuffer b, Triple t, String msg) {

    //    if(b.length() > 0)
      b.append('\n');
    if(t != null)
      b.append("INVALID TRIPLE: (" + t.subject() + ", " + t.predicate() + ", " + t.object() + "). ");
    b.append(msg);
  }

    
}
